package java.util.concurrent.atomic;

/**
 * An {@code int} value that may be updated atomically. See the
 * {@link java.util.concurrent.atomic} package specification for description of
 * the properties of atomic variables. An {@code AtomicInteger} is used in
 * applications such as atomically incremented counters, and cannot be used as a
 * replacement for an {@link java.lang.Integer}. However, this class does extend
 * {@code Number} to allow uniform access by tools and utilities that deal with
 * numerically-based classes.
 *
 * @since 1.5
 * @author Doug Lea
 */
public class AtomicInteger {
    private int value;

    /**
     * Creates a new AtomicInteger with initial value {@code 0}.
     */
    public AtomicInteger() {
    }

    /**
     * Creates a new AtomicInteger with the given initial value.
     *
     * @param initialValue
     *            the initial value
     */
    public AtomicInteger(int initialValue) {
        this.value = initialValue;
    }

    /**
     * Atomically sets the value to the given updated value if the current value
     * {@code ==} the expected value.
     *
     * @param expect
     *            the expected value
     * @param update
     *            the new value
     * @return {@code true} if successful. False return indicates that the
     *         actual value was not equal to the expected value.
     */
    public boolean compareAndSet(int expect, int update) {
        if (value == expect) {
            value = update;
            return true;
        } else {
            return false;
        }
    }

    /**
     * Gets the current value.
     *
     * @return the current value
     */
    public int get() {
        return value;
    }

    /**
     * Atomically increments by one the current value.
     *
     * @return the updated value
     */
    public int incrementAndGet() {
        return ++value;
    }

    /**
     * Sets to the given value.
     *
     * @param newValue
     *            the new value
     */
    public void set(int newValue) {
        value = newValue;
    }
}
