import mpi.Datatype;
import mpi.MPIException;
import mpi.Request;

//package mpi;

public class Intracomm /* extends Comm */ {
    /**
     * Size of group of this communicator.
     *
     * Java binding of the MPI operation <tt>MPI_COMM_SIZE</tt>.
     *
     * @return number of processors in the group of this communicator
     */
    public /*@ pure @*/ int Size() throws MPIException {
        return 0;
    }

    /*
     * Rank of this process in group of this communicator.
     *
     * Java binding of the MPI operation <tt>MPI_COMM_RANK</tt>.
     *
     * @return rank of the calling process in the group of this communicator
     */
    /*@ public pure int Rank() throws MPIException; @*/

    /**
     * Blocking send operation.
     * <p>
     * Java binding of the MPI operation <tt>MPI_SEND</tt>.
     * <p>
     * The actual argument associated with <tt>buf</tt> must be one-dimensional
     * array. The value <tt>offset</tt> is a subscript in this array, defining
     * the position of the first item of the message.
     * <p>
     * If the <tt>datatype</tt> argument represents an MPI basic type, its value
     * must agree with the element type of <tt>buf</tt>---either a primitive
     * type or a reference (object) type. If the <tt>datatype</tt> argument
     * represents an MPI derived type, its <em>base type</em> must agree with
     * the element type of <tt>buf</tt>
     *
     * @param buf
     *            send buffer array
     * @param offset
     *            initial offset in send buffer
     * @param count
     *            number of items to send
     * @param datatype
     *            datatype of each item in send buffer
     * @param dest
     *            rank of destination
     * @param tag
     *            message tag
     */
// begin(SendSpec)
    /*@
        given Future fut;
        given frac p;
        given process P;
        given frac bufp;
        requires fut != null ** p != none;
        requires buf.length == 1 ** offset == 0 ** count == 1;
        requires Perm(buf[0], bufp);
        requires valid(Rank(), dest, buf[0], tag);
        requires Future(fut, p, fut.send(Rank(), dest, buf[0], tag) * P);
        ensures Perm(buf[0], bufp);
        ensures Future(fut, p, P);
    @*/
    public void Send(Object buf, int offset, int count, Datatype datatype,
        int dest, int tag) throws MPIException {
        // end(SendSpec)
        Request req = Isend(buf, offset, count, datatype, dest, tag)
                /*@ with {fut = fut; p = p; P = fut.waitsend(r) * P; r = r;} @*/;
        Status s = req.Wait()/*@ with {fut = fut; p = p; P = P;} @*/;
        return s;
        // begin(SendSpec)
    }
// end(SendSpec)

    /**
     * Start a standard mode, nonblocking send.
     * <p>
     * Java binding of the MPI operation <tt>MPI_ISEND</tt>.
     * <p>
     * Further comments as for <tt>Send</tt>.
     *
     * @param buf
     *            send buffer array
     * @param offsetinitial
     *            offset in send buffer
     * @param count
     *            number of items to send
     * @param datatype
     *            datatype of each item in send buffer
     * @param dest
     *            rank of destination
     * @param tag
     *            message tag
     * @return communication request
     */
// begin(IsendSpec)
    /*@
        given Future fut;
        given frac p;
        given process P;
        given frac bufp;
        requires fut != null ** p != none;
        requires buf.length == 1 ** offset == 0 ** count == 1;
        requires Perm(buf[0], bufp);
        requires valid(Rank(), dest, buf[0], tag);
        requires Future(fut, p,
            \sum(int r; ; fut.isend(Rank(), dest, buf[0], tag, r) * P(r)));
        ensures Value(\result.recv) ** \result.recv == false **
            Value(\result.bufp) ** \result.bufp == bufp **
            Value(\result.buf) ** \result.buf = buf;
        ensures \result.waitToken();
        ensures Future(fut, p, P);
    @*/
    public Request Isend(Object buf, int offset, int count, Datatype datatype,
        int dest, int tag) throws MPIException {
        // end(IsendSpec)
        //@ action fut, p, P(r), fut.isend(source, dest, buf[0], tag, r);
        return new Request()
            /*@ with {recv = false; r = r; bufp = bufp; buf = buf;} @*/;
        // begin(IsendSpec)
    }
// end(IsendSpec)

    /**
     * Blocking receive operation.
     * <p>
     * Java binding of the MPI operation <tt>MPI_RECV</tt>.
     * <p>
     * The actual argument associated with <tt>buf</tt> must be one-dimensional
     * array. The value <tt>offset</tt> is a subscript in this array, defining
     * the position into which the first item of the incoming message will be
     * copied.
     * <p>
     * If the <tt>datatype</tt> argument represents an MPI basic type, its value
     * must agree with the element type of <tt>buf</tt>---either a primitive
     * type or a reference (object) type. If the <tt>datatype</tt> argument
     * represents an MPI derived type, its <em>base type</em> must agree with
     * the element type of <tt>buf</tt>
     *
     * @param buf
     *            receive buffer array
     * @param offset
     *            initial offset in receive buffer
     * @param count
     *            number of items in receive buffer
     * @param datatype
     *            datatype of each item in receive buffer
     * @param source
     *            rank of source
     * @param tag
     *            message tag
     * @return status object
     */
// begin(RecvSpec)
    /*@
        given Future fut;
        given frac p;
        given process P(msg);
        requires fut != null ** p != none;
        requires buf.length == 1 ** offset == 0 ** count == 1;
        requires Perm(buf[0], write);
        requires Future(fut, p,
            \sum(int act_source, int act_tag, Object msg;
                 this.source != ANY_SOURCE ==> act_source == this.source **
                 this.tag != ANY_TAG ==> act_tag == this.tag;
                 fut.recv(source, act_source, Rank(), msg, tag, act_tag, r) *
                 P(msg)));
        ensures Value(\result.source) ** \result.source == act_source;
        ensures Value(\result.tag) ** \result.tag == act_tag;
        ensures Perm(buf[0], write);
        ensures buf[0] == msg;
        ensures valid(act_source, Rank(), buf[0], act_tag);
        ensures Future(fut, p, P(msg));
    @*/
    public Status Recv(Object buf, int offset, int count, Datatype datatype,
            int source, int tag) throws MPIException {
        // end(RecvSpec)
        Request req = Irecv(buf, offset, count, datatype, source, tag)
                /*@ with {fut = fut; p = p; P = fut.wait(r) * P; r = r;} @*/;
        Status s = req.Wait()/*@ with {fut = fut; p = p; P(msg) = P(msg);} @*/;
        return s;
        // begin(RecvSpec)
    }
// end(RecvSpec)

    /**
     * Start a nonblocking receive.
     *
     * Java binding of the MPI operation <tt>MPI_IRECV</tt>.
     *
     * Further comments as for <tt>Recv</tt>.
     *
     * @param buf
     *            receive buffer array
     * @param offset
     *            initial offset in receive buffer
     * @param count
     *            number of items in receive buffer
     * @param datatype
     *            datatype of each item in receive buffer
     * @param source
     *            rank of source
     * @param tag
     *            message tag
     * @return communication request
     */
// begin(IrecvSpec)
    /*@
        given Future fut;
        given frac p;
        given process P(r);
        requires fut != null ** p != none;
        requires buf.length == 1 ** offset == 0 ** count == 1;
        requires Perm(buf[0], write);
        requires Future(fut, p,
            \sum(int r; ; fut.irecv(source, Rank(), tag, r) * P(r)));
        ensures Value(\result.recv) ** \result.recv == true **
            Value(\result.source) ** \result.source == source **
            Value(\result.dest) ** \result.dest == Rank() **
            Value(\result.tag) ** \result.tag == tag **
            Value(\result.r) ** \result.r == r **
            Value(\result.bufp) ** \result.bufp == write **
            Value(\result.buf) ** \result.buf = buf;
        ensures \result.waitToken();
        ensures Future(fut, p, P(r));
    @*/
    public Request Irecv(Object buf, int offset, int count, Datatype datatype, int source, int tag) throws MPIException {
        // end(IrecvSpec)
        //@ action fut, p, P(msg), fut.irecv(act_source, dest, act_tag, r);
        return new Request()
                /*@ with {recv = true; source = source; dest = Rank();
                          tag = tag; r = r; bufp = write; buf = buf;} @*/;
        // begin(IrecvSpec)
    }
// end(IrecvSpec)

    /**
     * A call to <tt>Barrier</tt> blocks the caller until all process in the
     * group have called it.
     *
     * Java binding of the MPI operation <tt>MPI_BARRIER</tt>.
     */
    public void Barrier() throws MPIException {
    }
}
