import java.io.Serializable;
import java.util.concurrent.atomic.AtomicInteger;

import mpi.MPI;
import mpi.Request;
import mpi.Status;

public class MPAO {
    /**
     * Counter used to generate unique call ids.
     */
    private static final AtomicInteger callIdCounter =
            new AtomicInteger(Integer.MAX_VALUE / 2);

    /**
     * The id of the Factory object.
     */
    public static final int FACTORY_ID = 0;

    private static FactoryRemote[] factories;

    public static void init() {
        int count = getWorkerCount();
        factories = new FactoryRemote[count];
        for (int i = 0; i < count; i++) {
            factories[i] = new FactoryRemote(i);
        }

        FactoryImpl localFactory = new FactoryImpl();
        localFactory.start();
    }

    public static FactoryRemote getFactory(int workerId) {
        return factories[workerId];
    }

    /**
     * Call a method on an active object.
     *
     * @param workerId
     *            The id of the MPI worker thread.
     * @param objectId
     *            The id of the active object.
     * @param signatue
     *            Method signature.
     * @param parameters
     *            Method parameters.
     * @return object representing the result of an active object method call.
     */
// begin(callMethod) begin(callMethodSpec)
    /*@
        given Future fut;
        given frac p;
        given process P(r);
        requires fut != null ** p != none;
        requires validCall(workerId, objectId, signature, parameters);
        requires Future(fut, p, 
            \sum(int callId, int r; ;
                fut.send(Rank(), workerId, new Call(callId, signature,
                                                    parameters), objectId) *
                          fut.irecv(workerId, Rank(), callId, r) * P(r)) *
            P);
        ensures \result.getToken();
        ensures Value(\result.request);
        ensures Value(\result.request.source) ** Value(\result.request.dest) **
            Value(\result.request.tag) ** Value(\result.request.r) **
            Value(\result.request.buf);
        ensures \result.request.source == workerId **
                \result.request.dest == Rank() **
                \result.request.tag == callId **
                \result.request.r == r **
                \result.request.bufp == write **
                \result.request.buf == \result.buffer;
        ensures Future(fut, p, P(r));
    @*/
    public static MPAOResult callMethod(int workerId, int objectId,
            String signature, Serializable[] parameters) {
        // end(callMethodSpec)
        // Call
        int callId = genCallId();
        Call call = new Call(callId, signature, parameters)
        /*@ with {dest = workerId; objectId = objectId;} @*/;
        
        //@ assert valid(Rank(), workerId, call, objectId);
        
        Object[] buf = new Object[] { call };
        MPI.COMM_WORLD.Send(buf, 0, buf.length, MPI.OBJECT, workerId,
                             objectId)
        /*@ with {fut = fut; p = p; bufp = write; P = 
            \sum(r; ; fut.recv(workerId, Rank(), tag, r) * P(r));}
        @*/;

        // Result
        Object[] resbuf = new Object[1];
        Request req = MPI.COMM_WORLD.Irecv(resbuf, 0, resbuf.length, MPI.OBJECT,
                                           workerId, callId)
        /*@ with {fut = fut; p = p; bufp = write; P = P(r);} @*/;

        return new MPAOResult(req, resbuf);
        // begin(callMethodSpec)
    }
// end(callMethod) end(callMethodSpec)

    /**
     * Generate an unique call id.
     *
     * Generates an unique identifier for an active object method call.
     *
     * @return Call id
     */
    private static int genCallId() {
        return callIdCounter.incrementAndGet();
    }

    /**
     * Receive a call to an active object method.
     *
     * @param objectId
     *            The id of the active object.
     * @return Active object method call information.
     */
// begin(receiveCall) begin(receiveCallSpec)
    /*@
        given Future fut;
        given frac p;
        given process P(msg);
        requires fut != null ** p != none;
        requires Future(fut, p,
            \sum(int source, Call msg; ;
            fut.recv(ANY_SOURCE, source, Rank(), msg, objectId, objectId, r) *
            P(msg)));
        ensures Value(\result.call) ** \result.call == msg;
        ensures Value(\result.source) ** \result.source == source;
        ensures \result.valid();
        ensures Future(fut, p, P(msg));
    @*/
    public static ReceivedCall receiveCall(int objectId) {
        // end(receiveCallSpec)
        Object[] buf = new Object[1];

        Status s = MPI.COMM_WORLD.Recv(buf, 0, buf.length, MPI.OBJECT,
                                       MPI.ANY_SOURCE, objectId)
        /*@ with {fut = fut; p = p; P(msg) = P(msg);} @*/;

        Call call = (Call) buf[0];
        ReceivedCall res = new ReceivedCall(s.source, call)
            /*@ with {objectId = objectId} @*/;

        return res;
        // begin(receiveCallSpec)
    }
// end(receiveCallSpec) end(receiveCall)

    /**
     * Get the number of workers.
     */
    public static int getWorkerCount() {
        return MPI.COMM_WORLD.Size();
    }

    /**
     * Get the worker id of the current worker.
     */
    public static int getWorkerId() {
        return MPI.COMM_WORLD.Rank();
    }

    /**
     * Send the result of an active object method call to the caller.
     *
     * @param call
     *            The call details.
     * @param result
     *            The result value associated with the call.
     */
// begin(sendResult) begin(sendResultSpec)
    /*@
        given Future fut;
        given frac p;
        given process P;
        requires fut != null ** p != none;
        requires postCondition(Rank(), call.objectId, call.signature,
            call.parameters, result);
        requires Future(fut, p, fut.send(Rank(), call.source,
            new Response(result), tag) * P);
        ensures Future(fut, p, P);
    @*/
    public static void sendResult(int source, Call call,
            Serializable result) {
        // end(sendResultSpec)
        Response resultObj = new Response(result) /*@ with {call = call;} @*/;
        Object[] buf = new Object[] { result };
        MPI.COMM_WORLD.Send(buf, 0, buf.length, MPI.OBJECT, source,
                             call.getCallId())
        /*@ with {fut = fut; p = p; bufp = write; P = P;} @*/;
        // begin(sendResultSpec)
    }
// end(sendResultSpec) end(sendResult)
}
